#!/usr/bin/env python
# coding: utf-8
import numpy as np
import matplotlib.pyplot as plt
import h5py
import scipy
import sklearn
import pydot


# 


def sigmoid(Z):
    """
    Implements the sigmoid activation in numpy
    
    Arguments:
    Z -- numpy array of any shape
    
    Returns:
    A -- output of sigmoid(z), same shape as Z
    cache -- returns Z as well, useful during backpropagation
    """
    
    A = 1/(1+np.exp(-Z))
    cache = Z
    
    return A, cache


# 


def tanh(Z):
    """
    Implements the tanh activation in numpy
    
    Arguments:
    Z -- numpy array of any shape
    
    Returns:
    A -- output of tanh(z), same shape as Z
    cache -- returns Z as well, useful during backpropagation
    """
    
    A = (np.exp(Z) - np.exp(-Z))/(np.exp(Z) + np.exp(-Z))
    cache = Z
    
    return A, cache


# 


def relu(Z):
    """
    Implement the RELU function.

    Arguments:
    Z -- Output of the linear layer, of any shape

    Returns:
    A -- Post-activation parameter, of the same shape as Z
    cache -- a python dictionary containing "A" ; stored for computing the backward pass efficiently
    """
    
    A = np.maximum(0,Z)
    
    assert(A.shape == Z.shape)
    
    cache = Z
    
    return A, cache


# 


def lrelu(Z):
    """
    Implement the 'leaky' RELU function.

    Arguments:
    Z -- Output of the linear layer, of any shape

    Returns:
    A -- Post-activation parameter, of the same shape as Z
    cache -- a python dictionary containing "A" ; stored for computing the backward pass efficiently
    """
    
    A = np.maximum(0.01*Z,Z)
    
    assert(A.shape == Z.shape)
    
    cache = Z 
    return A, cache


# 


def sigmoid_backward(dA, cache):
    """
    Implement the backward propagation for a single SIGMOID unit.

    Arguments:
    dA -- post-activation gradient, of any shape
    cache -- 'Z' where we store for computing backward propagation efficiently

    Returns:
    dZ -- Gradient of the cost with respect to Z
    """
    
    Z = cache
    
    s = 1/(1+np.exp(-Z))
    dZ = dA * s * (1-s)
    
    assert (dZ.shape == Z.shape)
    
    return dZ


# 


def tanh_backward(dA, cache):
    """
    Implement the backward propagation for a single TANH unit.

    Arguments:
    dA -- post-activation gradient, of any shape
    cache -- 'Z' where we store for computing backward propagation efficiently

    Returns:
    dZ -- Gradient of the cost with respect to Z
    """
    
    Z = cache
    
    s = (np.exp(Z) - np.exp(-Z))/(np.exp(Z) + np.exp(-Z))   
    dZ = dA * (1-s**2)
    
    assert (dZ.shape == Z.shape)
    
    return dZ


# 


def relu_backward(dA, cache):
    """
    Implement the backward propagation for a single RELU unit.

    Arguments:
    dA -- post-activation gradient, of any shape
    cache -- 'Z' where we store for computing backward propagation efficiently

    Returns:
    dZ -- Gradient of the cost with respect to Z
    """
    
    Z = cache
    dZ = np.array(dA, copy=True) # just converting dz to a correct object.
    
    # When z <= 0, you should set dz to 0 as well. 
    dZ[Z <= 0] = 0
    
    assert (dZ.shape == Z.shape)
    
    return dZ


# 


def lrelu_backward(dA, cache):
    """
    Implement the backward propagation for a single leaky RELU unit.

    Arguments:
    dA -- post-activation gradient, of any shape
    cache -- 'Z' where we store for computing backward propagation efficiently

    Returns:
    dZ -- Gradient of the cost with respect to Z
    """
    
    Z = cache
    dZ = np.array(dA, copy=True) # just converting dz to a correct object.
    
    # When z < 0, we set dz to 0.01 * dA.     
    dZ[Z == 0] = 0
    dZ[Z < 0] = 0.01 * dA
    
    assert (dZ.shape == Z.shape)
    
    return dZ


# 


def convert_to_one_hot(Y, C):
    Y = np.eye(C)[Y.reshape(-1)].T
    return Y


# 


def softmax(Z):
    """Compute the softmax of Z."""
    
    exps = np.exp(Z - np.max(Z, axis=0, keepdims=True))
    cache = Z    
    
    A = exps / np.sum(exps, axis=0, keepdims=True)
    
    return A, cache


# 


def compute_sofmax_cost(AL, Y):
    """
    Implement the cost function .

    Arguments:
    AL -- probability vector corresponding to the label predictions, shape (number of classes, number of examples)
    Y -- true "label" vector

    Returns:
    cost -- cross-entropy cost
    """
    
    m = Y.shape[1]

# Compute the average loss

    logprobs = -np.log(AL[Y.astype(bool)])
    print(logprobs)
    cost = np.sum(logprobs)/m
    
    # To make sure your cost's shape is what we expect.
    cost = np.squeeze(cost)      
    assert(cost.shape == ())
       
    return cost


# 


def compute_sofmax_dAL(AL,Y):
    """
    Calculate the derivative of the loss function with respect to A[L], dJ/dA[L] .

    Arguments:
    AL -- probability vector corresponding to the label predictions, shape (number of classes, number of examples)
    Y -- true "label" vector

    Returns:
    dAL -- derivative of the cost function with respect to AL
    """
    m = Y.shape[1]
    
    # Notice the scale factor (2/m) because J is the mean of all losses
    dAL = (2/m)*-1/AL 
    Y_zeros = ~Y.astype(bool)
    dAL[Y_zeros] = 0

    return dAL    


# 


def softmax_backward(dA, cache):
    """
    Implement the backward propagation for a single SOFTMAX unit.

    Arguments:
    dA -- post-activation gradient, of any shape
    cache -- 'Z' where we store for computing backward propagation efficiently

    Returns:
    dZ -- Gradient of the cost J with respect to Z
    """
    Z = cache
    A, activation_cache = softmax(Z)
    m = A.shape[1]
    
    dZ = np.zeros((dA.shape[0],m))
    
    dAdZ = np.zeros((dA.shape[0],dA.shape[0]))
    
    for k in range(m):
        A_col = A[:,k]
#         for i in range(len(A_col)):
#             for j in range(len(A_col)):
#                 if i == j:
#                     dAdZ[i,j] = A_col[i] * (1-A_col[i])
#                 else: 
#                     dAdZ[i,j] = -A_col[i] * A_col[j]
         
#     or more concisely using an outer product
        SM = A_col.reshape((-1,1))
        dAdZ = np.diag(A_col) - np.dot(SM, SM.T)

        dZ[:,k] = dAdZ@dA[:,k] 
    
    assert (dZ.shape == Z.shape)
    
    return dZ    


# 


def compute_dZL(AL,Y, cache):
    """
    Calculate the derivative of the loss function with respect to AL .

    Arguments:
    AL -- probability vector corresponding to the label predictions, shape (number of classes, number of examples)
    Y -- true "label" vector

    Returns:
    dAL -- derivative of the cost function with respect to AL
    """
    Z = cache    
    dZL = AL - Y
    
    return dZL    


# 


def numerical_dZL(Y,Z):
    step = 0.000001
    nrows,ncols = Y.shape
    cdZL = np.zeros((nrows,ncols))
    for col in range(ncols):
        Y_col = Y[:,col].copy().reshape(-1,1)
        # print(Y_col)

        for i in range(nrows):
            # print(Z)
            shift = np.zeros((nrows,1))
            shift[i] = step
            Z_plus = Z[:,col].copy()
            Z_plus = Z_plus.reshape(-1,1)
            Z_plus = Z_plus + shift
            # print(Z_plus)

            Z_minus = Z[:,col].copy()
            Z_minus = Z_minus.reshape(-1,1)
            Z_minus = Z_minus - shift    
            # print(Z_minus)

            A_plus, activation_cache = softmax(Z_plus)
            # print(A_plus,Y)
            A_minus, activation_cache = softmax(Z_minus)
            # print(A_minus,Y)
            J_plus = compute_sofmax_cost(A_plus, Y_col)
            J_minus = compute_sofmax_cost(A_minus, Y_col)
            cdZL[i,col] = (J_plus - J_minus)/(2*step)

    return cdZL


# 


def initialize_parameters_he(layers_dims):
    """
    Arguments:
    layer_dims -- python array (list) containing the size of each layer.
    
    Returns:
    parameters -- python dictionary containing your parameters "W1", "b1", ..., "WL", "bL":
                    W1 -- weight matrix of shape (layers_dims[1], layers_dims[0])
                    b1 -- bias vector of shape (layers_dims[1], 1)
                    ...
                    WL -- weight matrix of shape (layers_dims[L], layers_dims[L-1])
                    bL -- bias vector of shape (layers_dims[L], 1)
    """
    
    np.random.seed(1)
    parameters = {}
    nl = len(layers_dims) - 1 # integer representing the number of layers
     
    for l in range(1, nl):
        
        parameters['W' + str(l)] = np.random.randn(layers_dims[l],layers_dims[l-1])         * np.sqrt(2/layers_dims[l-1])
        parameters['b' + str(l)] = np.zeros((layers_dims[l],1))        

        assert(parameters['W' + str(l)].shape == (layer_dims[l], layer_dims[l-1]))
        assert(parameters['b' + str(l)].shape == (layer_dims[l], 1))        
        
    return parameters


# 


def initialize_parameters_xavier(layer_dims):
    """
    Arguments:
    layer_dims -- python array (list) containing the dimensions of each layer in the network
    
    Returns:
    parameters -- python dictionary containing your parameters "W1", "b1", ..., "WL", "bL":
                    Wl -- weight matrix of shape (layer_dims[l], layer_dims[l-1])
                    bl -- bias vector of shape (layer_dims[l], 1)
    """
    
    np.random.seed(1)
    parameters = {}          # we declare parameters as a "dictionary"
    nl = len(layer_dims)     # number of layers in the network including the input layer
       
    for l in range(1, nl):   # we skip layer [0], and the for loop count stops at nl-1
        
        parameters['W' + str(l)] = np.random.randn(layer_dims[l], layer_dims[l-1]) / np.sqrt(layer_dims[l-1]) # Alternatively multiply by 0.01
        parameters['b' + str(l)] = np.zeros((layer_dims[l], 1))
        
        
        assert(parameters['W' + str(l)].shape == (layer_dims[l], layer_dims[l-1]))
        assert(parameters['b' + str(l)].shape == (layer_dims[l], 1))
       
    return parameters


# 


def linear_forward(A, W, b):
    """
    Implement the linear part of a layer's forward propagation.

    Arguments:
    A -- activations from previous layer (or input data): (size of previous layer, number of examples)
    W -- weights matrix: numpy array of shape (size of current layer, size of previous layer)
    b -- bias vector, numpy array of shape (size of the current layer, 1)

    Returns:
    Z -- the input of the activation function, also called pre-activation parameter 
    cache -- a python "tuple" containing "A", "W" and "b" ; stored for computing the backward pass efficiently
    """
    
    Z = np.dot(W,A) + b
    
    # Alternatively we can use numpy newer syntax for matrix product
    # Z = W@A + b
       
    assert(Z.shape == (W.shape[0], A.shape[1]))
    cache = (A, W, b) # cache is a tuple
    
    return Z, cache


# 


def linear_activation_forward(A_prev, W, b, activation):
    """
    Implement the forward propagation for the LINEAR->ACTIVATION layer

    Arguments:
    A_prev -- activations from previous layer (or input data): (size of previous layer, number of examples)
    W -- weights matrix: numpy array of shape (size of current layer, size of previous layer)
    b -- bias vector, numpy array of shape (size of the current layer, 1)
    activation -- the activation to be used in this layer, stored as a text string: "sigmoid" or "relu"

    Returns:
    A -- the output (A[l]) of the activation function, also called the post-activation value 
    cache -- a python dictionary containing the "linear_cache" (A[l-1], W[l], b[l]) and
             the "activation_cache" (Z[l]);
             stored for computing the backward pass efficiently
    """
    
    if activation == "sigmoid":
        # Inputs: "A_prev, W, b". Outputs: "A, activation_cache".
        
        Z, linear_cache = linear_forward(A_prev, W, b) # linear cache is a tuple (A,W,b)
        A, activation_cache = sigmoid(Z)               # activation cache is Z

    elif activation == "softmax":
        # Inputs: "A_prev, W, b". Outputs: "A, activation_cache".
        
        Z, linear_cache = linear_forward(A_prev, W, b) # linear cache is a tuple (A,W,b)
        A, activation_cache = softmax(Z)                  # activation cache is Z                
    
    elif activation == "relu":
        # Inputs: "A_prev, W, b". Outputs: "A, activation_cache".
        
        Z, linear_cache = linear_forward(A_prev, W, b) # linear cache is a tuple (A,W,b)
        A, activation_cache = relu(Z)                  # activation cache is Z
        
    
    assert (A.shape == (W.shape[0], A_prev.shape[1]))
    cache = (linear_cache, activation_cache)           # cache is a tuple  

    return A, cache


# 


def L_model_forward(X, parameters, activation):
    """
    Implement forward propagation for the [LINEAR->RELU]*(L-1)->LINEAR->SIGMOID computation
    
    Arguments:
    X -- data, numpy array of shape (input size, number of examples)
    parameters -- output of initialize_parameters_deep()
    
    Returns:
    AL -- last post-activation value
    caches -- list of caches containing:
                every cache of linear_activation_forward() (there are L-1 of them, indexed from 0 to L-1)
    """

    caches = []  # caches is a "list" as opposed to cache, which is a "tuple"
    A = X
    L = len(parameters) // 2     # number of layers in the neural network
    
    # Implement [LINEAR -> RELU]*(L-1). Add "cache" to the "caches" list.
    for l in range(1, L):
        
        A_prev = A 
        
        A, cache = linear_activation_forward(A_prev, parameters['W' + str(l)], parameters['b' + str(l)], "relu")
        caches.append(cache)
        
    
    if activation == "sigmoid":
    # Implement LINEAR -> SIGMOID. Add "cache" to the "caches" list.
    
        AL, cache = linear_activation_forward(A, parameters['W' + str(L)], parameters['b' + str(L)], "sigmoid")
    elif activation == "softmax":
    # Implement LINEAR -> SOFTMAX. Add "cache" to the "caches" list.
        
        AL, cache = linear_activation_forward(A, parameters['W' + str(L)], parameters['b' + str(L)], "softmax")         

    # For debugging:
    # print(['W' + str(L) + ' dimensions = ' + str(parameters['W' + str(L)].shape)])
    # print(parameters['W' + str(L)].shape,X.shape[1])

    caches.append(cache)
    
    
    assert(AL.shape == (parameters['W' + str(L)].shape[0],X.shape[1]))
    
    return AL, caches
 


# 


def compute_cost(AL, Y, activation):
    """
    Implement the cost function defined by equation (33).

    Arguments:
    AL -- probability vector corresponding to your label predictions, shape (1, number of examples)
    Y -- true "label" vector (for example: containing 0 if non-cat, 1 if cat), shape (1, number of examples)

    Returns:
    cost -- cross-entropy cost
    """
    
    m = Y.shape[1]

    if activation == "sigmoid":
    # Compute loss from AL and y.    
        cost = (1./m) * (-np.dot(Y,np.log(AL).T) - np.dot(1-Y, np.log(1-AL).T))    

    elif activation == "softmax":
    # Compute the average loss from AL and y
        logprobs = -np.log(AL[Y.astype(bool)])
        cost = np.sum(logprobs)/m    
        
    # To make sure your cost's shape is what we expect (e.g. this turns [[22]] into 22).
    cost = np.squeeze(cost)      
    assert(cost.shape == ())
    
    return cost


# 


def linear_backward(dZ, cache):
    """
    Implement the linear portion of backward propagation for a single layer (layer l)

    Arguments:
    dZ -- Gradient of the cost with respect to the linear output (of current layer l)
    cache -- tuple of values (A_prev, W, b) coming from the forward propagation in the current layer

    Returns:
    dA_prev -- Gradient of the cost with respect to the activation (of the previous layer l-1), same shape as A_prev
    dW -- Gradient of the cost with respect to W (current layer l), same shape as W
    db -- Gradient of the cost with respect to b (current layer l), same shape as b
    """
    A_prev, W, b = cache
    m = A_prev.shape[1]

    
    dW = 1./m * np.dot(dZ,A_prev.T)
    db = 1./m * np.sum(dZ, axis = 1, keepdims = True)
    dA_prev = np.dot(W.T,dZ)
    
    
    assert (dA_prev.shape == A_prev.shape)
    assert (dW.shape == W.shape)
    assert (db.shape == b.shape)
    
    return dA_prev, dW, db


# 


def linear_activation_backward(dA, cache, activation):
    """
    Implement the backward propagation for the LINEAR->ACTIVATION layer.
    
    Arguments:
    dA -- post-activation gradient for current layer l 
    cache -- tuple of values (linear_cache, activation_cache) we store for computing backward propagation efficiently
    activation -- the activation to be used in this layer, stored as a text string: "sigmoid" or "relu"
    
    Returns:
    dA_prev -- Gradient of the cost with respect to the activation (of the previous layer l-1), same shape as A_prev
    dW -- Gradient of the cost with respect to W (current layer l), same shape as W
    db -- Gradient of the cost with respect to b (current layer l), same shape as b
    """
    linear_cache, activation_cache = cache
    
    if activation == "relu":
        
        dZ = relu_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ, linear_cache)
                
    elif activation == "sigmoid":
        
        dZ = sigmoid_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ, linear_cache)
            
    elif activation == "softmax":
        
        dZ = softmax_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward(dZ, linear_cache)
            
    return dA_prev, dW, db


# 


def L_model_backward(AL, Y, caches, activation):
    """
    Implement the backward propagation for the [LINEAR->RELU] * (L-1) -> LINEAR -> SIGMOID group
    
    Arguments:
    AL -- probability vector, output of the forward propagation (L_model_forward())
    Y -- true "label" vector (containing 0 if non-cat, 1 if cat)
    caches -- list of caches containing:
                every cache of linear_activation_forward() with "relu" (it's caches[l], for l in range(L-1) i.e l = 0...L-2)
                the cache of linear_activation_forward() with "sigmoid" (it's caches[L-1])
    
    Returns:
    grads -- A dictionary with the gradients
             grads["dA" + str(l)] = ... 
             grads["dW" + str(l)] = ...
             grads["db" + str(l)] = ... 
    """
    grads = {}      # this is a "dictionary"
    L = len(caches) # L is the number of layers
    
    m = AL.shape[1]
    Y = Y.reshape(AL.shape) # after this line, Y is the same shape as AL
    
    # Initializing the backpropagation
    if activation == "sigmoid":
        
        dAL = - np.divide(Y, AL) + np.divide(1 - Y, 1 - AL) # derivative of cost with respect to AL
            
    elif activation == "softmax":
    
        dAL = compute_sofmax_dAL(AL,Y)

    
    # Lth layer (SIGMOID/SOFTMAX -> LINEAR) gradients. Inputs: "dAL, current_cache". Outputs: "grads["dAL-1"], grads["dWL"], grads["dbL"]
    
    current_cache = caches[(L-1)]  # The index is L-1 because indices of the list start at 0
    grads["dA" + str(L-1)], grads["dW" + str(L)], grads["db" + str(L)] = linear_activation_backward(dAL, current_cache, activation)
    
    
    # Loop from l=L-2 to l=0
    for l in reversed(range(L-1)):
        # lth layer: (RELU -> LINEAR) gradients.
        # Inputs: "grads["dA" + str(l + 1)], current_cache". Outputs: "grads["dA" + str(l)] , grads["dW" + str(l + 1)] , grads["db" + str(l + 1)] 
        
        current_cache = caches[(l)]
        dA_prev_temp, dW_temp, db_temp = linear_activation_backward(grads["dA" + str(l + 1)], current_cache, activation = "relu")
        grads["dA" + str(l)] = dA_prev_temp
        grads["dW" + str(l + 1)] = dW_temp
        grads["db" + str(l + 1)] = db_temp
        

    return grads


# 


def update_parameters(parameters, grads, learning_rate):
    """
    Update parameters using gradient descent
    
    Arguments:
    parameters -- python dictionary containing your parameters 
    grads -- python dictionary containing your gradients, output of L_model_backward
    
    Returns:
    parameters -- python dictionary containing your updated parameters 
                  parameters["W" + str(l)] = ... 
                  parameters["b" + str(l)] = ...
    """
    
    L = len(parameters) // 2 # number of layers in the neural network

    # Update rule for each parameter. Use a for loop.
    
    for l in range(L):
        parameters["W" + str(l+1)] = parameters["W" + str(l+1)] - learning_rate*grads["dW" + str(l + 1)]
        parameters["b" + str(l+1)] = parameters["b" + str(l+1)] - learning_rate*grads["db" + str(l + 1)]
    
    return parameters


# 


def L_layer_model(X, Y, layers_dims, learning_rate = 0.0075, num_iterations = 3000, activation = "sigmoid", print_cost=False):#lr was 0.009
    """
    Implements a L-layer neural network: [LINEAR->RELU]*(L-1)->LINEAR->SIGMOID/SOFTMAX.
    
    Arguments:
    X -- data, numpy array of shape (number of examples, num_px * num_px * 3)
    Y -- true "label" vector (containing 0 if cat, 1 if non-cat), of shape (1, number of examples)
    layers_dims -- list containing the input size and each layer size, of length (number of layers + 1).
    learning_rate -- learning rate of the gradient descent update rule
    num_iterations -- number of iterations of the optimization loop
    print_cost -- if True, it prints the cost every 100 steps
    
    Returns:
    parameters -- parameters learnt by the model. They can then be used to predict.
    """

    np.random.seed(1)
    costs = []                         # keep track of cost
    
    # Parameters initialization. 
    parameters = initialize_parameters_xavier(layers_dims)
        
    # Loop (gradient descent)
    for i in range(0, num_iterations):

        # Forward propagation: [LINEAR -> RELU]*(L-1) -> LINEAR -> SIGMOID.        
        AL, caches = L_model_forward(X, parameters, activation)
        # print(AL.shape)
                
        # Compute cost.
        cost = compute_cost(AL, Y, activation)               
            
        # Backward propagation.        
        grads = L_model_backward(AL, Y, caches, activation)
         
        # Update parameters.        
        parameters = update_parameters(parameters, grads, learning_rate)
                        
        # Print the cost every 100 training example
        if print_cost and i % 100 == 0:
            print ("Cost after iteration %i: %f" %(i, cost))
        if print_cost and i % 100 == 0:
            costs.append(cost)
            
    # plot the cost
    plt.plot(np.squeeze(costs))
    plt.ylabel('cost')
    plt.xlabel('iterations (x 100)')
    plt.title("Learning rate =" + str(learning_rate))
    plt.show()
    
    return parameters


# 


def load_data():
    train_dataset = h5py.File('datasets/train_catvnoncat.h5', "r")
    train_set_x_orig = np.array(train_dataset["train_set_x"][:]) # your train set features
    train_set_y_orig = np.array(train_dataset["train_set_y"][:]) # your train set labels

    test_dataset = h5py.File('datasets/test_catvnoncat.h5', "r")
    test_set_x_orig = np.array(test_dataset["test_set_x"][:]) # your test set features
    test_set_y_orig = np.array(test_dataset["test_set_y"][:]) # your test set labels

    classes = np.array(test_dataset["list_classes"][:]) # the list of classes
    
    train_set_y_orig = train_set_y_orig.reshape((1, train_set_y_orig.shape[0]))
    test_set_y_orig = test_set_y_orig.reshape((1, test_set_y_orig.shape[0]))
    
    return train_set_x_orig, train_set_y_orig, test_set_x_orig, test_set_y_orig, classes


# 


def predict(X, Y, parameters, activation):
    """
    This function is used to predict the results of a  L-layer neural network.
    
    Arguments:
    X -- data set of examples you would like to label
    parameters -- parameters of the trained model
    
    Returns:
    p -- predictions for the given dataset X
    y_hat -- probabilities for the given dataset X
    """
    
    m = X.shape[1]
    n = len(parameters) // 2 # number of layers in the neural network
    nh = Y.shape[0]
    pred = np.zeros((nh,m)) # these are the predictions
    
    # Forward propagation: y_hat are the probabilities
    AL, caches = L_model_forward(X, parameters, activation)

    if activation == "softmax":    
    # convert probabilities to 0/1 predictions
        for i in range(0, AL.shape[1]):
            ind = np.unravel_index(np.argmax(AL[:,i], axis=None), AL[:,i].shape)
            pred[ind[0],i] = 1

        a = pred + Y
        correct_indices = np.asarray(np.where(a == 2))[0,:].reshape(-1,1)
        accuracy = len(correct_indices)/m
            
    elif activation == "sigmoid": 
        
#         for i in range(0, AL.shape[1]):
#             if AL[0,i] > 0.5:
#                 pred[0,i] = 1
#             else:
#                 pred[0,i] = 0
                 
        pred = np.rint(AL).astype(int)
        accuracy = np.sum((pred == Y)/m)
        
    return pred, AL, accuracy


# 


def find_mislabeled_images(classes, X, y, p):
    """
    Plots images where predictions and truth were different.
    X -- dataset
    y -- true labels
    p -- predictions
    """
    a = p + y
    print(a)
    # print(p.shape,y.shape,a.shape)
    mislabeled_indices = np.asarray(np.where(a[0,:] == 1))
    # print(mislabeled_indices.shape)
    num_images = len(mislabeled_indices[0])
    # mislabeled_indices = mislabeled_indices[1,:].reshape(num_images,1)
    
    return mislabeled_indices, num_images 


# 


def L_layer_model_with_costs(X, Y, X_test, Y_test, layers_dims, learning_rate = 0.0075,                              num_iterations = 3000, activation = "sigmoid", save_cost=True, print_cost=False):
    """
    Implements a L-layer neural network: [LINEAR->RELU]*(L-1)->LINEAR->SIGMOID.
    
    Arguments:
    X -- data, numpy array of shape (number of examples, num_px * num_px * 3)
    Y -- true "label" vector (containing 0 if cat, 1 if non-cat), of shape (1, number of examples)
    layers_dims -- list containing the input size and each layer size, of length (number of layers + 1).
    learning_rate -- learning rate of the gradient descent update rule
    num_iterations -- number of iterations of the optimization loop
    print_cost -- if True, it prints the cost every 100 steps
    
    Returns:
    parameters -- parameters learnt by the model. They can then be used to predict.
    """

    np.random.seed(1)
   
    costs = [] # keep track of cost
    costs_test = [] # keep track of cost in the test set
    accu_test = [] # keep tract of accuracy on the test set
    
    # Parameters initialization. 
    parameters = initialize_parameters_xavier(layers_dims)
        
    # Loop (gradient descent)
    for i in range(0, num_iterations):

        # Forward propagation: [LINEAR -> RELU]*(L-1) -> LINEAR -> SIGMOID.        
        AL, caches = L_model_forward(X, parameters, activation)    
        
        # Compute cost.
        cost = compute_cost(AL, Y, activation)
                                                                           
        # Backward propagation.        
        grads = L_model_backward(AL, Y, caches, activation)
         
        # Update parameters.        
        parameters = update_parameters(parameters, grads, learning_rate)
        
        # Print the cost every 100 training example
        if save_cost and i % 100 == 0:
            pred_test, AL_test, acc_test = predict(X_test, Y_test, parameters, activation)
            cost_test = compute_cost(AL_test, Y_test, activation)
            costs.append(np.squeeze(cost).tolist())
            costs_test.append(np.squeeze(cost_test).tolist())
            accu_test.append(np.squeeze(acc_test).tolist())
            if print_cost:
                print ("Train Set Cost,  Test Set Cost, Test Set Accuracy, after iteration %i: %f %f %f" \
                       %(i, cost, cost_test, acc_test))
    
    return parameters, costs, costs_test, accu_test


# 


def compute_cost_with_L2_regularization(AL, Y, parameters, lambd, activation):
    """
    Implement the cost function with L2 regularization. See formula (2) above.
    
    Arguments:
    A3 -- post-activation, output of forward propagation, of shape (output size, number of examples)
    Y -- "true" labels vector, of shape (output size, number of examples)
    parameters -- python dictionary containing parameters of the model
    
    Returns:
    cost - value of the regularized loss function (formula (2))
    """
    m = Y.shape[1]
  
    L = len(parameters) // 2 # number of layers in the neural network

    # Use a for loop to go through all the weight matrices
    
    regularization_cost = 0.0
    
    for l in range(L):

         regularization_cost += (lambd/(2*m))*np.sum(np.square(parameters["W" + str(l+1)]))
   
    cross_entropy_cost = compute_cost(AL, Y, activation) # This gives you the cross-entropy part of the cost
        
    cost = cross_entropy_cost + regularization_cost
    
    return cost


# 


def L_model_backward_with_L2_regularization(AL, Y, caches, lambd, activation):
    """
    Implement the backward propagation for the [LINEAR->RELU] * (L-1) -> LINEAR -> SIGMOID group
    
    Arguments:
    AL -- probability vector, output of the forward propagation (L_model_forward())
    Y -- true "label" vector (containing 0 if non-cat, 1 if cat)
    caches -- list of caches containing:
                every cache of linear_activation_forward() with "relu" (it's caches[l], for l in range(L-1) i.e l = 0...L-2)
                the cache of linear_activation_forward() with "sigmoid" (it's caches[L-1])
    
    Returns:
    grads -- A dictionary with the gradients
             grads["dA" + str(l)] = ... 
             grads["dW" + str(l)] = ...
             grads["db" + str(l)] = ... 
    """
    grads = {}      # this is a "dictionary"
    L = len(caches) # L is the number of layers
    
    m = AL.shape[1]
    Y = Y.reshape(AL.shape) # after this line, Y is the same shape as AL
    
    # Initializing the backpropagation
    if activation == "sigmoid":
        
        dAL = - np.divide(Y, AL) + np.divide(1 - Y, 1 - AL) # derivative of cost with respect to AL
            
    elif activation == "softmax":
    
        dAL = compute_sofmax_dAL(AL,Y)
    
    # Lth layer (SIGMOID -> LINEAR) gradients. Inputs: "dAL, current_cache". Outputs: "grads["dAL-1"], grads["dWL"], grads["dbL"]
    
    current_cache = caches[(L-1)]  # The index is L-1 because indices of the list start at 0
    grads["dA" + str(L-1)], grads["dW" + str(L)], grads["db" + str(L)] = linear_activation_backward_with_L2_regularization(dAL, current_cache, lambd, activation)
    
    
    # Loop from l=L-2 to l=0
    for l in reversed(range(L-1)):
        # lth layer: (RELU -> LINEAR) gradients.
        # Inputs: "grads["dA" + str(l + 1)], current_cache". Outputs: "grads["dA" + str(l)] , grads["dW" + str(l + 1)] , grads["db" + str(l + 1)] 
        
        current_cache = caches[(l)]
        dA_prev_temp, dW_temp, db_temp = linear_activation_backward_with_L2_regularization(grads["dA" + str(l + 1)], current_cache, lambd, activation = "relu")
        grads["dA" + str(l)] = dA_prev_temp
        grads["dW" + str(l + 1)] = dW_temp
        grads["db" + str(l + 1)] = db_temp
        

    return grads


# 


def linear_activation_backward_with_L2_regularization(dA, cache, lambd, activation):
    """
    Implement the backward propagation for the LINEAR->ACTIVATION layer.
    
    Arguments:
    dA -- post-activation gradient for current layer l 
    cache -- tuple of values (linear_cache, activation_cache) we store for computing backward propagation efficiently
    activation -- the activation to be used in this layer, stored as a text string: "sigmoid" or "relu"
    
    Returns:
    dA_prev -- Gradient of the cost with respect to the activation (of the previous layer l-1), same shape as A_prev
    dW -- Gradient of the cost with respect to W (current layer l), same shape as W
    db -- Gradient of the cost with respect to b (current layer l), same shape as b
    """
    linear_cache, activation_cache = cache
    
    if activation == "relu":
        
        dZ = relu_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward_with_L2_regularization(dZ, linear_cache, lambd)
                
    elif activation == "sigmoid":
        
        dZ = sigmoid_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward_with_L2_regularization(dZ, linear_cache, lambd)

    elif activation == "softmax":
        
        dZ = softmax_backward(dA,activation_cache)
        dA_prev, dW, db = linear_backward_with_L2_regularization(dZ, linear_cache, lambd)

    return dA_prev, dW, db


# 


def linear_backward_with_L2_regularization(dZ, cache, lambd):
    """
    Implement the linear portion of backward propagation for a single layer (layer l)

    Arguments:
    dZ -- Gradient of the cost with respect to the linear output (of current layer l)
    cache -- tuple of values (A_prev, W, b) coming from the forward propagation in the current layer

    Returns:
    dA_prev -- Gradient of the cost with respect to the activation (of the previous layer l-1), same shape as A_prev
    dW -- Gradient of the cost with respect to W (current layer l), same shape as W
    db -- Gradient of the cost with respect to b (current layer l), same shape as b
    """
    A_prev, W, b = cache
    m = A_prev.shape[1]

    
    dW = 1./m * np.dot(dZ,A_prev.T) + lambd*W/m
    db = 1./m * np.sum(dZ, axis = 1, keepdims = True)
    dA_prev = np.dot(W.T,dZ)
    
    
    assert (dA_prev.shape == A_prev.shape)
    assert (dW.shape == W.shape)
    assert (db.shape == b.shape)
    
    return dA_prev, dW, db


# 


def L_layer_model_with_L2_regularization(X, Y, X_test, Y_test, layers_dims, lambd,                                         \
                                         learning_rate, num_iterations = 3000,activation = "sigmoid", \
                                         save_cost=True, print_cost=False):
    """
    Implements a L-layer neural network: [LINEAR->RELU]*(L-1)->LINEAR->SIGMOID (or SOFTMAX).
    
    Arguments:
    X -- data, numpy array of shape (number of examples, num_px * num_px * 3)
    Y -- true "label" vector (containing 0 if cat, 1 if non-cat), of shape (1, number of examples)
    layers_dims -- list containing the input size and each layer size, of length (number of layers + 1).
    learning_rate -- learning rate of the gradient descent update rule
    num_iterations -- number of iterations of the optimization loop
    print_cost -- if True, it prints the cost every 100 steps
    
    Returns:
    parameters -- parameters learnt by the model. They can then be used to predict.
    """

    np.random.seed(1)
   
    costs = [] # keep track of cost
    costs_test = []
    accu_test = []
    
    # Parameters initialization. 
    parameters = initialize_parameters_xavier(layers_dims)
        
    # Loop (gradient descent)
    for i in range(0, num_iterations):

        # Forward propagation: [LINEAR -> RELU]*(L-1) -> LINEAR -> SIGMOID.        
        AL, caches = L_model_forward(X, parameters, activation)
        
        # Compute cost.        
        cost = compute_cost_with_L2_regularization(AL, Y, parameters, lambd, activation)
                                                                                           
        # Backward propagation.        
        grads = L_model_backward_with_L2_regularization(AL, Y, caches, lambd, activation)
         
        # Update parameters.        
        parameters = update_parameters(parameters, grads, learning_rate)
                        
        # Print the cost every 100 training example
        if save_cost and i % 100 == 0:
            pred_test, AL_test, acc_test = predict(X_test, Y_test, parameters, activation)
            # cost_test = compute_cost_with_L2_regularization(AL_test, Y_test, parameters, lambd, activation)
            cost_test = compute_cost(AL_test, Y_test, activation)
            costs.append(np.squeeze(cost).tolist())
            costs_test.append(np.squeeze(cost_test).tolist())
            accu_test.append(np.squeeze(acc_test).tolist())
            if print_cost:
                print ("Train Set Cost,  Test Set Cost, Test Set Accuracy, after iteration %i: %f %f %f" \
                       %(i, cost, cost_test, acc_test))
                
    return parameters, costs, costs_test, accu_test


# 


def linear_activation_forward_with_dropout(A_prev, W, b, keep_prob, activation):
    """
    Implement the forward propagation with dropout for the LINEAR->ACTIVATION layer. Notice that the activations "sigmoid" 
    and "softmax" of this function in practice are not used. 
    
    Arguments:
    A_prev -- activations from previous layer (or input data): (size of previous layer, number of examples)
    W -- weights matrix: numpy array of shape (size of current layer, size of previous layer)
    b -- bias vector, numpy array of shape (size of the current layer, 1)
    activation -- the activation to be used in this layer, stored as a text string: "sigmoid" or "relu"

    Returns:
    A -- the output (A[l]) of the activation function, also called the post-activation value 
    cache -- a python dictionary containing the "linear_cache" (A[l-1], W[l], b[l]) and
             the "activation_cache" (Z[l]);
             stored for computing the backward pass efficiently
    """
    
    if activation == "sigmoid":
        # Inputs: "A_prev, W, b". Outputs: "A, activation_cache".
        
        Z, linear_cache = linear_forward(A_prev, W, b) # linear cache is a tuple (A,W,b)
        A, activation_cache = sigmoid(Z)               # activation cache is Z
        
    elif activation == "softmax":
        # Inputs: "A_prev, W, b". Outputs: "A, activation_cache".
        
        Z, linear_cache = linear_forward(A_prev, W, b) # linear cache is a tuple (A,W,b)
        A, activation_cache = softmax(Z)               # activation cache is Z
            
    elif activation == "relu":
        # Inputs: "A_prev, W, b". Outputs: "A, activation_cache".
        
        Z, linear_cache = linear_forward(A_prev, W, b) # linear cache is a tuple (A,W,b)
        A, activation_cache = relu(Z)                  # activation cache is Z
        
        # Here we select the neurons to shut down
        
        D = np.random.rand(A.shape[0],A.shape[1])      # Step 1: initialize matrix D1 = np.random.rand(..., ...)
        D = D < keep_prob                              # Step 2: convert entries of D1 to 0 or 1 (using keep_prob as the threshold)
        A = A * D                                      # Step 3: shut down some neurons of A1
        A = A / keep_prob  
        cache_2 = D                                    # cache_2 is the matrix D
    
    assert (A.shape == (W.shape[0], A_prev.shape[1]))
    cache = (linear_cache, activation_cache)           # cache is a tuple  

    return A, cache, cache_2


# 


def L_model_forward_with_dropout(X, parameters, keep_prob, activation):
    """
    Implement forward propagation for the [LINEAR->RELU]*(L-1)->LINEAR->SIGMOID computation
    
    Arguments:
    X -- data, numpy array of shape (input size, number of examples)
    parameters -- output of initialize_parameters_deep()
    
    Returns:
    AL -- last post-activation value
    caches -- list of caches containing:
                every cache of linear_activation_forward() (there are L-1 of them, indexed from 0 to L-1)
    """

    caches = []   # caches is a "list" as opposed to cache, which is a "tuple"
    caches_2 = [] # caches_2 is a "list" as opposed to cache_2, which is a "matrix"
    A = X
    L = len(parameters) // 2     # number of layers in the neural network
    
    # Implement [LINEAR -> RELU]*(L-1). Add "cache" to the "caches" list.
    for l in range(1, L):
        
        A_prev = A 
        
        A, cache, cache_2 = linear_activation_forward_with_dropout(A_prev, parameters['W' + str(l)],                                              parameters['b' + str(l)], keep_prob, "relu")
        caches.append(cache)
        caches_2.append(cache_2)
        
    
    # Implement LINEAR -> SIGMOID. Add "cache" to the "caches" list.
    
    AL, cache = linear_activation_forward(A, parameters['W' + str(L)], parameters['b' + str(L)], activation)
    
    # print(AL)
    caches.append(cache)
    
    
    assert(AL.shape == (parameters['W' + str(L)].shape[0],X.shape[1]))
    
    return AL, caches, caches_2
 

# 


def L_model_backward_with_L2_regularization_and_dropout(AL, Y, caches, caches_2, lambd, keep_prob, activation):
    """
    Implement the backward propagation for the [LINEAR->RELU] * (L-1) -> LINEAR -> SIGMOID group
    
    Arguments:
    AL -- probability vector, output of the forward propagation (L_model_forward())
    Y -- true "label" vector (containing 0 if non-cat, 1 if cat)
    caches -- list of caches containing:
                every cache of linear_activation_forward() with "relu" (it's caches[l], for l in range(L-1) i.e l = 0...L-2)
                the cache of linear_activation_forward() with "sigmoid" (it's caches[L-1])
    
    Returns:
    grads -- A dictionary with the gradients
             grads["dA" + str(l)] = ... 
             grads["dW" + str(l)] = ...
             grads["db" + str(l)] = ... 
    """
    grads = {}      # this is a "dictionary"
    L = len(caches) # L is the number of layers
    
    m = AL.shape[1]
    Y = Y.reshape(AL.shape) # after this line, Y is the same shape as AL
    
    # Initializing the backpropagation
    if activation == "sigmoid":    
        dAL = - np.divide(Y, AL) + np.divide(1 - Y, 1 - AL) # derivative of cost with respect to AL
    elif activation == "softmax":
        dAL = compute_sofmax_dAL(AL,Y)    

    # Lth layer (SIGMOID -> LINEAR) gradients. Inputs: "dAL, current_cache". Outputs: "grads["dAL-1"], grads["dWL"], grads["dbL"]
    
    current_cache = caches[(L-1)]  # The index is L-1 because indices of the list start at 0
    grads["dA" + str(L-1)], grads["dW" + str(L)], grads["db" + str(L)] = linear_activation_backward_with_L2_regularization(dAL, current_cache, lambd, activation)
    # print(grads["dA" + str(L-1)].shape, caches_2[L-2].shape)
    grads["dA" + str(L-1)] = (grads["dA" + str(L-1)] * caches_2[L-2]) / keep_prob
    
    # Loop from l=L-2 to l=0
    for l in reversed(range(L-1)):
        # lth layer: (RELU -> LINEAR) gradients.
        # Inputs: "grads["dA" + str(l + 1)], current_cache". Outputs: "grads["dA" + str(l)] , grads["dW" + str(l + 1)] , grads["db" + str(l + 1)] 
        # print(l)
        current_cache = caches[(l)]
        D_prev = caches_2[l-1]
        dA_prev_temp, dW_temp, db_temp = linear_activation_backward_with_L2_regularization(grads["dA" + str(l + 1)], current_cache, lambd, activation = "relu")
        # print(dA_prev_temp.shape,D_prev.shape)
        if l > 0:
            grads["dA" + str(l)] = (dA_prev_temp * D_prev)/keep_prob
        elif l == 0:
            grads["dA" + str(l)] = dA_prev_temp
        grads["dW" + str(l + 1)] = dW_temp
        grads["db" + str(l + 1)] = db_temp
        

    return grads


# 


def L_layer_model_with_L2_regularization_and_dropout(X, Y, X_test, Y_test, layers_dims, \
                                                     lambd, keep_prob, learning_rate, num_iterations = 3000, \
                                                     activation = "Sigmoid", \
                                                     save_cost=True, print_cost=False, print_D=False): 
    """
    Implements a L-layer neural network: [LINEAR->RELU]*(L-1)->LINEAR->SIGMOID.
    
    Arguments:
    X -- data, numpy array of shape (number of examples, num_px * num_px * 3)
    Y -- true "label" vector (containing 0 if cat, 1 if non-cat), of shape (1, number of examples)
    layers_dims -- list containing the input size and each layer size, of length (number of layers + 1).
    learning_rate -- learning rate of the gradient descent update rule
    num_iterations -- number of iterations of the optimization loop
    print_cost -- if True, it prints the cost every 100 steps
    
    Returns:
    parameters -- parameters learnt by the model. They can then be used to predict.
    """

    np.random.seed(1)

    costs = [] # keep track of cost
    costs_test = [] # keep track of cost in the test set
    accu_test = [] # keep track of accuracy in the test set
    
    # Parameters initialization. 
    parameters = initialize_parameters_xavier(layers_dims)
        
    # Loop (gradient descent)
    for i in range(0, num_iterations):

        # Forward propagation: [LINEAR -> RELU]*(L-1) -> LINEAR -> SIGMOID.        
        AL, caches, caches_2 = L_model_forward_with_dropout(X, parameters, keep_prob, activation)
        
        # Show D matrices
        if print_D and i % 100 == 0:
            len_D = len(caches_2)
            # print(caches_2[0].shape)
            # print(caches_2[1].shape)
            # print(caches_2[2].shape)    
            fig, axes = plt.subplots(1, len_D, figsize=(16, 8))
            for i_axes in range(0, len_D):
                axes[i_axes].imshow(caches_2[i_axes])
                axes[i_axes].set_xlabel("D" + str(i_axes+1))
            plt.show()
        
        # Correct for underflow and overflow in the cost function
        AL_0 = AL == 0
        AL_1 = AL == 1
        AL[AL_0] = 0 + 1e-09
        AL[AL_1] = 1 - 1e-09
                
        # if np.sum(AL_0) > 0:
        #     print("AL_0 sum > 0")
        #     print(AL_0)
        #     print(AL)
            
        # elif np.sum(AL_1) > 0:
        #     print("AL_1 sum > 0")
        #     print(AL_1)
        #    print(AL)
            
        # Compute cost.        
        cost = compute_cost_with_L2_regularization(AL, Y, parameters, lambd, activation)
                                                                                                           
        # Backward propagation.        
        grads = L_model_backward_with_L2_regularization_and_dropout(AL, Y, caches, caches_2, lambd, keep_prob, activation)
         
        # Update parameters.        
        parameters = update_parameters(parameters, grads, learning_rate)
                        
        # Print the cost every 100 training example
        if save_cost and i % 100 == 0:
            pred_test, AL_test, acc_test = predict(X_test, Y_test, parameters, activation)
            # cost_test = compute_cost_with_L2_regularization(AL_test, Y_test, parameters, lambd, activation)
            cost_test = compute_cost(AL_test, Y_test, activation)
            costs.append(np.squeeze(cost).tolist())
            costs_test.append(np.squeeze(cost_test).tolist())
            accu_test.append(np.squeeze(acc_test).tolist())
            if print_cost:
                print ("Train Set Cost,  Test Set Cost, Test Set Accuracy, after iteration %i: %f %f %f" \
                       %(i, cost, cost_test, acc_test).tolist())
                
    return parameters, costs, costs_test, accu_test

