#!/usr/bin/env python
# coding: utf-8

# # Residual Networks
# 
# In this notebook we learn how to build a very deep convolutional network. In theory, very deep networks can represent very complex functions; but in practice, they are hard to train. Residual Networks, introduced by [He et al.](https://arxiv.org/pdf/1512.03385.pdf), allow the training of much deeper networks than previously practically feasible.
# 
# We will build a ResNet using ***Keras*** with its default ***TensorFlow*** backend in two steps: 
# 1. We will implement the basic building blocks of ResNets. 
# 2. We will put together these building blocks to implement and train a state-of-the-art neural network for image classification. 
# 
# 
# As usual, we start by loading the required packages.

# In[1]:


import numpy as np
from keras import layers
from keras.layers import Input, Add, Dense, Activation, ZeroPadding2D, BatchNormalization, Flatten, Conv2D, AveragePooling2D, MaxPooling2D, GlobalMaxPooling2D
from keras.models import Model, load_model
from keras.preprocessing import image
from keras.utils import layer_utils
from keras.utils.data_utils import get_file
from keras.applications.imagenet_utils import preprocess_input
import pydot
from IPython.display import SVG
from keras.utils.vis_utils import model_to_dot
from keras.utils import plot_model
from NB5_utils import *
from keras.initializers import glorot_uniform
import scipy.misc
from matplotlib.pyplot import imshow
get_ipython().run_line_magic('matplotlib', 'inline')

import keras.backend as K
K.set_image_data_format('channels_last')
K.set_learning_phase(1)


# ## The problem with very deep neural networks
# 
# In recent years, neural networks have become deeper, with state-of-the-art networks going from just a few layers (e.g., AlexNet) to over a hundred layers.
# 
# The main benefit of a very deep network is that it can represent very complex functions. It can also learn features at many different levels of abstraction, from edges (at the lower layers) to very complex features (at the deeper layers). However, using a deeper network doesn't always help. A huge barrier to training them is vanishing gradients: very deep networks often have a gradient signal that goes to zero quickly, thus making gradient descent unbearably slow. More specifically, during gradient descent, as you backprop from the final layer back to the first layer, you are multiplying by the weight matrix on each step, and thus the gradient can decrease exponentially quickly to zero (or, in rare cases, grow exponentially quickly and "explode" to take very large values). 
# 
# During training, you might therefore see the magnitude (or norm) of the gradient for the earlier layers descrease to zero very rapidly as training proceeds: 
# 
# <img src="images/NB5_IMG_1.png" style="width:450px;height:220px;">
# 
# Furthermore, it is often observed that as the depth of the network increases, the performance (% accuracy on the training set) increases up to a point and then starts decreasing again. This is a consequence of the difficulty of refining a very large number of parameters and the progressive accumulation of training errors in these parameters.   
# 
# These common problems of *deep neural networks* are largely resolved by the introduction of ***Residual Networks*** or ***ResNets***. *Resnets* are built on the following idea. Imagine you are calculating the activations of 2 fully connected layers in a NN:   
# .   
# 
# <center><img src = "images/NB5_IMG_2.png"></center>   
# 
# To flow from $a^{[l]}$ to $a^{[l+2]}$ information has to follow a long path. Now, imagine we add a second shorter path for $a^{[l]}$ to rapidly flow downstream:   
# .   
# 
# <center><img src = "images/NB5_IMG_3.png"></center>   
# 
# This is now called a **_Residual Block_**, and a **_Residual Network_** is obtained by stacking together several of these *residual blocks*.   
# 
# We can understand why using *residual blocks* eliminates the accumulation of parameters errors during the training of a NN by looking at the following example. Imagine we have a *"Big NN"* producing some activation $a^{[l]}$ from an input $x$, and we observe that adding two extra layers to the networks, thus producing a new activation $a^{[l+2]}$, degrades its accuracy instead of increasing it. We decide to replace the two extra layers with a *residual block*, such that now we have:   
# .   
# 
# <center><img src = "images/NB5_IMG_4.png"></center> 
# 
# $$a^{[l+2]}=g(z^{[l+2]}+a^{[l]})$$   
# 
# $$a^{[l+2]}=g(w^{[l+2]}a^{[l+1]}+b^{[l+2]}+a^{[l]})$$   
# 
# 
# Notice that if $a^{[l+2]}$ and $a^{[l]}$ do not have the same dimensions, and:   
# 
# $$a^{[l+2]}\in {{\mathbb{R}}^{m\times 1}}$$   
# $$a^{[l]}\in {{\mathbb{R}}^{n\times 1}}$$   
# 
# the expression above is corrected to:   
# 
# $$a^{[l+2]}=g(w^{[l+2]}a^{[l+1]}+b^{[l+2]}+W_{s}a^{[l]})$$   
# 
# where $W_{s}\in {{\mathbb{R}}^{m\times n}}$.   
# 
# We recall here that if we use **_L2 regularization_** this corresponds to progressively shrinking the magnitude of $w$ and $b$ (*L2 reg.* = *weight decay*). If $w^{[l+2]}a^{[l+1]}+b^{[l+2]}$ tends to 0, then:   
# 
# $$a^{[l+2]}=g(w^{[l+2]}a^{[l+1]}+b^{[l+2]}+a^{[l]})=g(a^{[l]})$$   
# 
# and if we are using a **ReLU** function for activation, such that $a\ge0$:   
# 
# $$g(a^{[l]})=a^{[l]}$$
# 
# $$a^{[l+2]}=a^{[l]}$$   
# 
# This means that if addition of the extra 2 layers improves the accuracy of the network, that's great, but if it does not, the residual block can rapidly learn the **_identity function_**, and propagate $a^{[l]}$ forward without degradation of the network accuracy. The figure below, taken from the original RESNET paper [He et al., 2015. Deep residual networks for image recognition] show how a traditional *plain* convolutional network is converted into a *residual network* by replacing standard blocks with residual blocks. Most convolutions are of the *same* type, with exception of a few residual blocks in which a $W_{s}$ matrix is used to correct the mismatch of dimensions between $a^{[l+2]}$ and $a^{[l]}$:    
# 
# <center><img src = "images/NB5_IMG_5.png"></center>   
# 

# ## Building a Residual Network
# 
# In ResNets a "shortcut" or a "skip connection" allows the gradient to be directly backpropagated to earlier layers. We also saw that having ResNet blocks with the shortcut makes it very easy for one of the blocks to learn an identity function. This means that we can stack on additional ResNet blocks with little risk of harming training set performance. There is also some evidence that the ease of learning an identity function more than skip connections helps with vanishing gradients, and accounts for ResNets' remarkable performance.
# 
# Two main types of blocks are used in a ResNet, depending mainly on whether the input/output dimensions are same or different.  

# ### The identity block
# 
# The identity block is the standard block used in ResNets, and corresponds to the case where the input activation (say $a^{[l]}$) has the same dimension as the output activation (say $a^{[l+2]}$). To flesh out the different steps of what happens in a ResNet's identity block, here is an alternative diagram showing the individual steps:
# 
# <img src="images/NB5_IMG_6.png" style="width:650px;height:150px;">
# 
# The upper path is the "shortcut path." The lower path is the "main path." In this diagram, we have also made explicit the CONV2D and ReLU steps in each layer. To speed up training we have also added a BatchNorm step. Don't worry about this being complicated to implement--you'll see that BatchNorm is just one line of code in Keras! 
# 
# Here, we actually implement a slightly more powerful version of this identity block, in which the skip connection "skips over" 3 hidden layers rather than 2 layers. It looks like this: 
# 
# <img src="images/NB5_IMG_7.png" style="width:650px;height:150px;">
# 
# 
# These are the individual steps.
# 
# First component of main path: 
# - The first CONV2D has $F_1$ filters of shape (1,1) and a stride of (1,1). Its padding is "valid" and its name should be `conv_name_base + '2a'`. Use 0 as the seed for the random initialization. 
# - The first BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '2a'`.
# - Then apply the ReLU activation function. This has no name and no hyperparameters. 
# 
# Second component of main path:
# - The second CONV2D has $F_2$ filters of shape $(f,f)$ and a stride of (1,1). Its padding is "same" and its name should be `conv_name_base + '2b'`. Use 0 as the seed for the random initialization. 
# - The second BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '2b'`.
# - Then apply the ReLU activation function. This has no name and no hyperparameters. 
# 
# Third component of main path:
# - The third CONV2D has $F_3$ filters of shape (1,1) and a stride of (1,1). Its padding is "valid" and its name should be `conv_name_base + '2c'`. Use 0 as the seed for the random initialization. 
# - The third BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '2c'`. Note that there is no ReLU activation function in this component. 
# 
# Final step: 
# - The shortcut and the input are added together.
# - Then apply the ReLU activation function. This has no name and no hyperparameters. 
# 
# It is important to read the implementation carefully to understand what it is doing. The following are additional links that help in this process:   
# 
# - To implement the Conv2D step: [See reference](https://keras.io/layers/convolutional/#conv2d)
# - To implement BatchNorm: [See reference](https://faroit.github.io/keras-docs/1.2.2/layers/normalization/) (axis: Integer, the axis that should be normalized (typically the channels axis))
# - For the activation, use:  `Activation('relu')(X)`
# - To add the value passed forward by the shortcut: [See reference](https://keras.io/layers/merge/#add)   
# 

# In[2]:


def identity_block(X, f, filters, stage, block):
    """
    Implementation of the identity block as defined in Figure 3
    
    Arguments:
    X -- input tensor of shape (m, n_H_prev, n_W_prev, n_C_prev)
    f -- integer, specifying the shape of the middle CONV's window for the main path
    filters -- python list of integers, defining the number of filters in the CONV layers of the main path
    stage -- integer, used to name the layers, depending on their position in the network
    block -- string/character, used to name the layers, depending on their position in the network
    
    Returns:
    X -- output of the identity block, tensor of shape (n_H, n_W, n_C)
    """
    
    # defining name basis
    conv_name_base = 'res' + str(stage) + block + '_branch'
    bn_name_base = 'bn' + str(stage) + block + '_branch'
    
    # Retrieve Filters
    F1, F2, F3 = filters
    
    # Save the input value. You'll need this later to add back to the main path. 
    X_shortcut = X
    
    # First component of main path
    X = Conv2D(filters = F1, kernel_size = (1, 1), strides = (1,1), padding = 'valid', name = conv_name_base                + '2a', kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = bn_name_base + '2a')(X)
    X = Activation('relu')(X)
    
    # Second component of main path 
    X = Conv2D(filters = F2, kernel_size = (f, f), strides = (1,1), padding = 'same', name = conv_name_base                + '2b', kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = bn_name_base + '2b')(X)
    X = Activation('relu')(X)

    # Third component of main path 
    X = Conv2D(filters = F3, kernel_size = (1, 1), strides = (1,1), padding = 'valid', name = conv_name_base                + '2c', kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = bn_name_base + '2c')(X)

    # Final step: Add shortcut value to main path, and pass it through a RELU activation 
    X = layers.Add()([X_shortcut, X])
    X = Activation('relu')(X)
    
    return X


# In[3]:


tf.reset_default_graph()

with tf.Session() as test:
    np.random.seed(1)
    A_prev = tf.placeholder("float", [3, 4, 4, 6])
    X = np.random.randn(3, 4, 4, 6)
    A = identity_block(A_prev, f = 2, filters = [2, 4, 6], stage = 1, block = 'a')
    test.run(tf.global_variables_initializer())
    out = test.run([A], feed_dict={A_prev: X, K.learning_phase(): 0})
    print("out = " + str(out[0][1][1][0]))


# ### The convolutional block
# 
# The ResNet "convolutional block" is the other type of block. We use this type of block when the input and output dimensions don't match up. The difference with the identity block is that there is a CONV2D layer in the shortcut path: 
# 
# <img src="images/NB5_IMG_7b.png" style="width:650px;height:150px;">
# 
# The CONV2D layer in the shortcut path is used to resize the input $x$ to a different dimension, so that the dimensions match up in the final addition needed to add the shortcut value back to the main path. (This plays a similar role as the matrix $W_s$ discussed in lecture.) For example, to reduce the activation dimensions's height and width by a factor of 2, you can use a 1x1 convolution with a stride of 2. The CONV2D layer on the shortcut path does not use any non-linear activation function. Its main role is to just apply a (learned) linear function that reduces the dimension of the input, so that the dimensions match up for the later addition step. 
# 
# The details of the convolutional block are as follows. 
# 
# First component of main path:
# - The first CONV2D has $F_1$ filters of shape (1,1) and a stride of (s,s). Its padding is "valid" and its name should be `conv_name_base + '2a'`. 
# - The first BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '2a'`.
# - Then apply the ReLU activation function. This has no name and no hyperparameters. 
# 
# Second component of main path:
# - The second CONV2D has $F_2$ filters of (f,f) and a stride of (1,1). Its padding is "same" and it's name should be `conv_name_base + '2b'`.
# - The second BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '2b'`.
# - Then apply the ReLU activation function. This has no name and no hyperparameters. 
# 
# Third component of main path:
# - The third CONV2D has $F_3$ filters of (1,1) and a stride of (1,1). Its padding is "valid" and it's name should be `conv_name_base + '2c'`.
# - The third BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '2c'`. Note that there is no ReLU activation function in this component. 
# 
# Shortcut path:
# - The CONV2D has $F_3$ filters of shape (1,1) and a stride of (s,s). Its padding is "valid" and its name should be `conv_name_base + '1'`.
# - The BatchNorm is normalizing the channels axis.  Its name should be `bn_name_base + '1'`. 
# 
# Final step: 
# - The shortcut and the main path values are added together.
# - Then apply the ReLU activation function. This has no name and no hyperparameters. 
#     
# Below we have implement the convolutional block. It is important to read the implementation carefully to understand what it is doing. The following are additional links that help in this process:    
# 
# - [Conv Hint](https://keras.io/layers/convolutional/#conv2d)
# - [BatchNorm Hint](https://keras.io/layers/normalization/#batchnormalization) (axis: Integer, the axis that should be normalized (typically the features axis))
# - For the activation, use:  `Activation('relu')(X)`
# - [Addition Hint](https://keras.io/layers/merge/#add)

# In[4]:


def convolutional_block(X, f, filters, stage, block, s = 2):
    """
    Implementation of the convolutional block as defined in Figure 4
    
    Arguments:
    X -- input tensor of shape (m, n_H_prev, n_W_prev, n_C_prev)
    f -- integer, specifying the shape of the middle CONV's window for the main path
    filters -- python list of integers, defining the number of filters in the CONV layers of the main path
    stage -- integer, used to name the layers, depending on their position in the network
    block -- string/character, used to name the layers, depending on their position in the network
    s -- Integer, specifying the stride to be used
    
    Returns:
    X -- output of the convolutional block, tensor of shape (n_H, n_W, n_C)
    """
    
    # defining name basis
    conv_name_base = 'res' + str(stage) + block + '_branch'
    bn_name_base = 'bn' + str(stage) + block + '_branch'
    
    # Retrieve Filters
    F1, F2, F3 = filters
    
    # Save the input value
    X_shortcut = X


    ##### MAIN PATH #####
    # First component of main path 
    X = Conv2D(F1, kernel_size = (1, 1), strides = (s,s), padding = 'valid', name = conv_name_base + '2a',                kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = bn_name_base + '2a')(X)
    X = Activation('relu')(X)

    # Second component of main path 
    X = Conv2D(F2, kernel_size = (f, f), strides = (1,1), padding = 'same', name = conv_name_base + '2b',                kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = bn_name_base + '2b')(X)
    X = Activation('relu')(X)

    # Third component of main path 
    X = Conv2D(F3, kernel_size = (1, 1), strides = (1,1), padding = 'valid', name = conv_name_base + '2c',                kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = bn_name_base + '2c')(X)

    ##### SHORTCUT PATH #### 
    X_shortcut = Conv2D(F3, kernel_size = (1, 1), strides = (s,s), padding = 'valid', name = conv_name_base + '1',                kernel_initializer = glorot_uniform(seed=0))(X_shortcut)
    X_shortcut = BatchNormalization(axis = 3, name = bn_name_base + '1')(X_shortcut)

    # Final step: Add shortcut value to main path, and pass it through a RELU activation
    X = layers.Add()([X_shortcut, X])
    X = Activation('relu')(X)
    
    
    return X


# In[5]:


tf.reset_default_graph()

with tf.Session() as test:
    np.random.seed(1)
    A_prev = tf.placeholder("float", [3, 4, 4, 6])
    X = np.random.randn(3, 4, 4, 6)
    A = convolutional_block(A_prev, f = 2, filters = [2, 4, 6], stage = 1, block = 'a')
    test.run(tf.global_variables_initializer())
    out = test.run([A], feed_dict={A_prev: X, K.learning_phase(): 0})
    print("out = " + str(out[0][1][1][0]))


# ## Building a ResNet model with 50 layers
# 
# We now have the necessary blocks to build a very deep ResNet. The following figure describes in detail the architecture of this neural network. "ID BLOCK" in the diagram stands for "Identity block," and "ID BLOCK x3" means you should stack 3 identity blocks together.
# 
# <img src="images/NB5_IMG_8.png" style="width:850px;height:150px;">
# 
# 
# The details of this ResNet-50 model are:
# - Zero-padding pads the input with a pad of (3,3)
# - Stage 1:
#     - The 2D Convolution has 64 filters of shape (7,7) and uses a stride of (2,2). Its name is "conv1".
#     - BatchNorm is applied to the channels axis of the input.
#     - MaxPooling uses a (3,3) window and a (2,2) stride.
# - Stage 2:
#     - The convolutional block uses three set of filters of size [64,64,256], "f" is 3, "s" is 1 and the block is "a".
#     - The 2 identity blocks use three set of filters of size [64,64,256], "f" is 3 and the blocks are "b" and "c".
# - Stage 3:
#     - The convolutional block uses three set of filters of size [128,128,512], "f" is 3, "s" is 2 and the block is "a".
#     - The 3 identity blocks use three set of filters of size [128,128,512], "f" is 3 and the blocks are "b", "c" and "d".
# - Stage 4:
#     - The convolutional block uses three set of filters of size [256, 256, 1024], "f" is 3, "s" is 2 and the block is "a".
#     - The 5 identity blocks use three set of filters of size [256, 256, 1024], "f" is 3 and the blocks are "b", "c", "d", "e" and "f".
# - Stage 5:
#     - The convolutional block uses three set of filters of size [512, 512, 2048], "f" is 3, "s" is 2 and the block is "a".
#     - The 2 identity blocks use three set of filters of size [512, 512, 2048], "f" is 3 and the blocks are "b" and "c".
# - The 2D Average Pooling uses a window of shape (2,2) and its name is "avg_pool".
# - The flatten doesn't have any hyperparameters or name.
# - The Fully Connected (Dense) layer reduces its input to the number of classes using a softmax activation. Its name should be `'fc' + str(classes)`.
# 
# Below we have implement the ResNet with 50 layers described in the figure above. It uses the following functions:
# 
# 
# - Average pooling [see reference](https://keras.io/layers/pooling/#averagepooling2d)
# - Conv2D: [See reference](https://keras.io/layers/convolutional/#conv2d)
# - BatchNorm: [See reference](https://keras.io/layers/normalization/#batchnormalization) (axis: Integer, the axis that should be normalized (typically the features axis))
# - Zero padding: [See reference](https://keras.io/layers/convolutional/#zeropadding2d)
# - Max pooling: [See reference](https://keras.io/layers/pooling/#maxpooling2d)
# - Fully conected layer: [See reference](https://keras.io/layers/core/#dense)
# - Addition: [See reference](https://keras.io/layers/merge/#add)

# In[6]:


def ResNet50(input_shape = (64, 64, 3), classes = 6):
    """
    Implementation of the popular ResNet50 the following architecture:
    CONV2D -> BATCHNORM -> RELU -> MAXPOOL -> CONVBLOCK -> IDBLOCK*2 -> CONVBLOCK -> IDBLOCK*3
    -> CONVBLOCK -> IDBLOCK*5 -> CONVBLOCK -> IDBLOCK*2 -> AVGPOOL -> TOPLAYER

    Arguments:
    input_shape -- shape of the images of the dataset
    classes -- integer, number of classes

    Returns:
    model -- a Model() instance in Keras
    """
    
    # Define the input as a tensor with shape input_shape
    X_input = Input(input_shape)

    
    # Zero-Padding
    X = ZeroPadding2D((3, 3))(X_input)
    
    # Stage 1
    X = Conv2D(64, (7, 7), strides = (2, 2), name = 'conv1', kernel_initializer = glorot_uniform(seed=0))(X)
    X = BatchNormalization(axis = 3, name = 'bn_conv1')(X)
    X = Activation('relu')(X)
    X = MaxPooling2D((3, 3), strides=(2, 2))(X)

    # Stage 2
    X = convolutional_block(X, f = 3, filters = [64, 64, 256], stage = 2, block='a', s = 1)
    X = identity_block(X, 3, [64, 64, 256], stage=2, block='b')
    X = identity_block(X, 3, [64, 64, 256], stage=2, block='c')

    # Stage 3 
    X = convolutional_block(X, f = 3, filters = [128, 128, 512], stage = 3, block='a', s = 2)
    X = identity_block(X, 3, [128, 128, 512], stage=3, block='b')
    X = identity_block(X, 3, [128, 128, 512], stage=3, block='c')
    X = identity_block(X, 3, [128, 128, 512], stage=3, block='d')

    # Stage 4 
    X = convolutional_block(X, f = 3, filters = [256, 256, 1024], stage = 4, block='a', s = 2)
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block='b')
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block='c')
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block='d')
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block='e')
    X = identity_block(X, 3, [256, 256, 1024], stage=4, block='f')

    # Stage 5 
    X = convolutional_block(X, f = 3, filters = [512, 512, 2048], stage = 5, block='a', s = 2)
    X = identity_block(X, 3, [512, 512, 2048], stage=5, block='b')
    X = identity_block(X, 3, [512, 512, 2048], stage=5, block='c')

    # AVGPOOL. Use "X = AveragePooling2D(...)(X)"
    X = AveragePooling2D(pool_size=(2, 2),name = 'avg_pool')(X)    

    # Output layer
    X = Flatten()(X)
    X = Dense(classes, activation='softmax', name='fc' + str(classes), kernel_initializer = glorot_uniform(seed=0))(X)
        
    # Create model
    model = Model(inputs = X_input, outputs = X, name='ResNet50')

    return model


# Run the following code to build the model's graph. If your implementation is not correct you will know it by checking your accuracy when running `model.fit(...)` below.

# In[7]:


model = ResNet50(input_shape = (64, 64, 3), classes = 6)


# As seen in the Keras Tutorial Notebook, prior training a model, you need to configure the learning process by compiling the model.

# In[8]:


model.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])


# The model is now ready to be trained. The only thing you need is a dataset.

# Let's load the SIGNS Dataset.
# 
# <img src="images/NB5_IMG_9.png" style="width:450px;height:250px;">
# 

# In[7]:


X_train_orig, Y_train_orig, X_test_orig, Y_test_orig, classes = load_dataset()

# Normalize image vectors
X_train = X_train_orig/255.
X_test = X_test_orig/255.

# Print Y_train
print(Y_train_orig)
print(Y_train_orig.reshape(-1))

# Convert training and test labels to one hot matrices
Y_train = convert_to_one_hot(Y_train_orig, 6).T
Y_test = convert_to_one_hot(Y_test_orig, 6).T

print ("number of training examples = " + str(X_train.shape[0]))
print ("number of test examples = " + str(X_test.shape[0]))
print ("X_train shape: " + str(X_train.shape))
print ("Y_train shape: " + str(Y_train.shape))
print ("X_test shape: " + str(X_test.shape))
print ("Y_test shape: " + str(Y_test.shape))


# We run the following cell to train the model on 2 epochs with a batch size of 32. On a CPU it should take you around 5min per epoch. 

# In[10]:


model.fit(X_train, Y_train, epochs = 2, batch_size = 32)


# Let's see how this model (trained on only two epochs) performs on the test set.

# In[11]:


preds = model.evaluate(X_test, Y_test)
print ("Loss = " + str(preds[0]))
print ("Test Accuracy = " + str(preds[1]))


# In the cell above we have trained the model only for two epochs, and thus it achieves poor performance. ResNet50 is a powerful model for image classification when it is trained for an adequate number of iterations. We would get a much better performance if we trained for ~40 epochs, but this would take hours when training on a CPU. Alternatively, model's weights for ResNet50 trained for 20 epochs using the SIGNS dataset are also available in the *models* directory.

# In[10]:


model.fit(X_train, Y_train, epochs = 40, batch_size = 32)

# %%
 
model.save('models/ResNet50_2epochs.h5')  # creates a HDF5 file 'ResNet50_2epochs.h5'


# In[11]:


preds = model.evaluate(X_test, Y_test)
print ("Loss = " + str(preds[0]))
print ("Test Accuracy = " + str(preds[1]))


# We can now save the model in the models directory:

# In[17]:


model.save('models/ResNet50_40epochs.h5')  # creates a HDF5 file 'ResNet50_40epochs.h5'


# In[ ]:


# del model  # deletes the existing model
# model = load_model('models/ResNet50_40epochs.h5') 
# preds = model.evaluate(X_test, Y_test)
# print ("Loss = " + str(preds[0]))
# print ("Test Accuracy = " + str(preds[1]))


# #### Test an image not in the test set.

# We can take a picture of your own hand and see the output of the model. To do this:
#     1. Click on "File" in the upper bar of this notebook, then click "Open" to go on your Coursera Hub.
#     2. Add your image to this Jupyter Notebook's directory, in the "images" folder
#     3. Write your image's name in the following code
#     4. Run the code and check if the algorithm is right! 

# In[12]:


img_path = 'images/my_image.jpg'
img = image.load_img(img_path, target_size=(64, 64))
x = image.img_to_array(img)
x = np.expand_dims(x, axis=0)
x = preprocess_input(x)
print('Input image shape:', x.shape)
my_image = scipy.misc.imread(img_path)
imshow(my_image)
print("class prediction vector [p(0), p(1), p(2), p(3), p(4), p(5)] = ")
print(model.predict(x))


# We can print a summary of the ResNet model.

# In[13]:


model.summary()


# Finally, we can run the code below to visualize the ResNet50 model we have created. A .png picture of your model is also available as  images/model.png".

# In[14]:

import pydot
plot_model(model, to_file='ResNet50_model.ps')
SVG(model_to_dot(model).create(prog='dot', format='svg'))


# ### References 
# 
# This notebook presents the ResNet algorithm due to He et al. (2015) following the implementation deposited in the github repository by Francois Chollet: 
# 
# - Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun - [Deep Residual Learning for Image Recognition (2015)](https://arxiv.org/abs/1512.03385)
# - Francois Chollet's github repository: https://github.com/fchollet/deep-learning-models/blob/master/resnet50.py   
# 
# 
# Additional notebooks on NN implemetation with Keras/Python can be found at the github repository  https://github.com/fchollet/deep-learning-with-python-notebooks , which contains all the jupyter notebooks accompanying Chollet's book  **_Deep Learning with Python_** (https://www.manning.com/books/deep-learning-with-python?a_aid=keras&a_bid=76564dff). Note that the original text of the book features far more content than you will find in these notebooks, in particular further explanations and figures.
# 
# 

# In[ ]:




